from collections.abc import Awaitable, Callable
from typing import TypeVar

from returns.interfaces.specific.future import FutureLikeN
from returns.primitives.hkt import Kinded, KindN, kinded

_FirstType = TypeVar('_FirstType')
_SecondType = TypeVar('_SecondType')
_ThirdType = TypeVar('_ThirdType')
_UpdatedType = TypeVar('_UpdatedType')

_FutureKind = TypeVar('_FutureKind', bound=FutureLikeN)


def bind_awaitable(
    function: Callable[[_FirstType], Awaitable[_UpdatedType]],
) -> Kinded[
    Callable[
        [KindN[_FutureKind, _FirstType, _SecondType, _ThirdType]],
        KindN[_FutureKind, _UpdatedType, _SecondType, _ThirdType],
    ]
]:
    """
    Composes a container a regular ``async`` function.

    This function should return plain, non-container value.

    In other words, it modifies the function's
    signature from:
    ``a -> Awaitable[b]``
    to:
    ``Container[a] -> Container[b]``

    This is how it should be used:

    .. code:: python

        >>> import anyio
        >>> from returns.future import Future
        >>> from returns.io import IO
        >>> from returns.pointfree import bind_awaitable

        >>> async def coroutine(x: int) -> int:
        ...    return x + 1

        >>> assert anyio.run(
        ...     bind_awaitable(coroutine)(Future.from_value(1)).awaitable,
        ... ) == IO(2)

    Note, that this function works
    for all containers with ``.bind_awaitable`` method.
    See :class:`returns.primitives.interfaces.specific.future.FutureLikeN`
    for more info.

    """

    @kinded
    def factory(
        container: KindN[_FutureKind, _FirstType, _SecondType, _ThirdType],
    ) -> KindN[_FutureKind, _UpdatedType, _SecondType, _ThirdType]:
        return container.bind_awaitable(function)

    return factory



from abc import abstractmethod
from collections.abc import Callable
from typing import Generic, Never, TypeVar

from returns.primitives.hkt import KindN

_FirstType = TypeVar('_FirstType')
_SecondType = TypeVar('_SecondType')
_ThirdType = TypeVar('_ThirdType')
_UpdatedType = TypeVar('_UpdatedType')

_BindableType = TypeVar('_BindableType', bound='BindableN')


class BindableN(Generic[_FirstType, _SecondType, _ThirdType]):
    """
    Represents a "context" in which calculations can be executed.

    ``Bindable`` allows you to bind together
    a series of calculations while maintaining
    the context of that specific container.

    In contrast to :class:`returns.interfaces.lashable.LashableN`,
    works with the first type argument.
    """

    __slots__ = ()

    @abstractmethod
    def bind(
        self: _BindableType,
        function: Callable[
            [_FirstType],
            KindN[_BindableType, _UpdatedType, _SecondType, _ThirdType],
        ],
    ) -> KindN[_BindableType, _UpdatedType, _SecondType, _ThirdType]:
        """
        Applies 'function' to the result of a previous calculation.

        And returns a new container.
        """


#: Type alias for kinds with one type argument.
Bindable1 = BindableN[_FirstType, Never, Never]

#: Type alias for kinds with two type arguments.
Bindable2 = BindableN[_FirstType, _SecondType, Never]

#: Type alias for kinds with three type arguments.
Bindable3 = BindableN[_FirstType, _SecondType, _ThirdType]



"""
Typing ``pipe`` functions requires several phases.

It is pretty obvious from its usage:

1. When we pass a sequence of functions we have to reduce
   the final callable type, it is require to match the ``callable`` protocol.
   And at this point we also kinda try
   to check that all pipeline functions do match,
   but this is impossible to do 100% correctly at this point,
   because generic functions don't have a type argument
   to infer the final result

2. When we call the function, we need to check for two things.
   First, we check that passed argument fits our instance requirement.
   Second, we check that pipeline functions match.
   Now we have all arguments to do the real inference.

3. We also need to fix generic in method signature.
   It might be broken, because we add new generic arguments and return type.
   So, it is safe to reattach generic back to the function.

Here's when it works:

.. code:: python

  >>> from returns.pipeline import pipe

  >>> def first(arg: int) -> bool:
  ...     return arg > 0
  >>> def second(arg: bool) -> str:
  ...     return 'bigger' if arg else 'not bigger'

  >>> pipeline = pipe(first, second)  # `analyzed` is called
  >>> assert pipeline(1) == 'bigger'  # `signature and `infer` are called
  >>> assert pipeline(0) == 'not bigger'  # `signature and `infer` again

"""

from collections.abc import Callable

from mypy.nodes import ARG_POS
from mypy.plugin import FunctionContext, MethodContext, MethodSigContext
from mypy.types import (
    AnyType,
    CallableType,
    FunctionLike,
    Instance,
    ProperType,
    TypeOfAny,
    UnionType,
    get_proper_type,
    get_proper_types,
)
from mypy.types import Type as MypyType

from returns.contrib.mypy._typeops.analtype import translate_to_function
from returns.contrib.mypy._typeops.inference import PipelineInference
from returns.contrib.mypy._typeops.transform_callable import detach_callable


def analyze(ctx: FunctionContext) -> MypyType:
    """This hook helps when we create the pipeline from sequence of funcs."""
    default_return = get_proper_type(ctx.default_return_type)
    if not isinstance(default_return, Instance):
        return default_return

    if not ctx.arg_types[0]:  # We do require to pass `*functions` arg.
        ctx.api.fail('Too few arguments for "pipe"', ctx.context)
        return default_return

    arg_types = [arg_type[0] for arg_type in ctx.arg_types if arg_type]
    first_step, last_step = _get_pipeline_def(arg_types, ctx)
    if not isinstance(first_step, FunctionLike):
        return default_return
    if not isinstance(last_step, FunctionLike):
        return default_return

    return default_return.copy_modified(
        args=[
            # First type argument represents first function arguments type:
            _unify_type(first_step, _get_first_arg_type),
            # Second argument represents pipeline final return type:
            _unify_type(last_step, lambda case: case.ret_type),
            # Other types are just functions inside the pipeline:
            *arg_types,
        ],
    )


def infer(ctx: MethodContext) -> MypyType:
    """This hook helps when we finally call the created pipeline."""
    if not isinstance(ctx.type, Instance):
        return ctx.default_return_type

    pipeline_functions = get_proper_types(ctx.type.args[2:])
    return PipelineInference(
        get_proper_type(ctx.arg_types[0][0]),
    ).from_callable_sequence(
        pipeline_functions,
        list((ARG_POS,) * len(pipeline_functions)),
        ctx,
    )


def signature(ctx: MethodSigContext) -> CallableType:
    """Helps to fix generics in method signature."""
    return detach_callable(ctx.default_signature)


def _get_first_arg_type(case: CallableType) -> MypyType:
    """Function might not have args at all."""
    if case.arg_types:
        return case.arg_types[0]
    return AnyType(TypeOfAny.implementation_artifact)


def _unify_type(
    function: FunctionLike,
    fetch_type: Callable[[CallableType], MypyType],
) -> MypyType:
    return UnionType.make_union([fetch_type(case) for case in function.items])


def _get_pipeline_def(
    arg_types: list[MypyType],
    ctx: FunctionContext,
) -> tuple[ProperType, ProperType]:
    first_step = get_proper_type(arg_types[0])
    last_step = get_proper_type(arg_types[-1])

    if not isinstance(first_step, FunctionLike):
        first_step = translate_to_function(first_step, ctx)
    if not isinstance(last_step, FunctionLike):
        last_step = translate_to_function(last_step, ctx)
    return first_step, last_step



import pickle  # noqa: S403

from returns.primitives.container import BaseContainer


def test_pickle_backward_deserialization():
    """Test that BaseContainer can be deserialized from 0.19.0 and earlier."""
    # BaseContainer(1) serialized as of 0.19.0
    serialized_container = (
        b'\x80\x04\x958\x00\x00\x00\x00\x00\x00\x00\x8c\x1c'
        + b'returns.primitives.container\x94\x8c\rBaseContainer'
        + b'\x94\x93\x94)\x81\x94K\x01b.'
    )
    assert pickle.loads(serialized_container) == BaseContainer(1)  # noqa: S301



from returns.result import Failure, Success


def test_map_success():
    """Ensures that Success is mappable."""
    assert Success(5).map(str) == Success('5')


def test_alt_failure():
    """Ensures that Failure is mappable."""
    assert Failure(5).map(str) == Failure(5)
    assert Failure(5).alt(str) == Failure('5')


def test_alt_success():
    """Ensures that Success.alt is NoOp."""
    assert Success(5).alt(str) == Success(5)



import pytest

from returns.context import (
    RequiresContextFutureResult,
    RequiresContextIOResult,
    RequiresContextResult,
)
from returns.contrib.pytest import ReturnsAsserts
from returns.functions import identity
from returns.future import FutureResult
from returns.io import IOFailure, IOSuccess
from returns.result import Failure, Success


def _under_test(
    container,
    *,
    should_lash: bool = False,
):
    if should_lash:
        return container.lash(container.from_failure)
    return container.bind(container.from_value)


@pytest.mark.parametrize(
    'container',
    [
        Success(1),
        Failure(1),
        IOSuccess(1),
        IOFailure(1),
        RequiresContextIOResult.from_value(1),
        RequiresContextIOResult.from_failure(1),
        RequiresContextFutureResult.from_value(1),
        RequiresContextFutureResult.from_failure(1),
        RequiresContextResult.from_value(1),
        RequiresContextResult.from_failure(1),
        FutureResult.from_value(1),
        FutureResult.from_failure(1),
    ],
)
@pytest.mark.parametrize(
    'kwargs',
    [
        {'should_lash': True},
    ],
)
def test_error_handled(returns: ReturnsAsserts, container, kwargs):
    """Demo on how to use ``pytest`` helpers to work with error handling."""
    assert not returns._errors_handled  # noqa: SLF001
    error_handled = _under_test(container, **kwargs)

    assert returns.is_error_handled(error_handled)
    assert returns.is_error_handled(error_handled.map(identity))
    assert returns.is_error_handled(error_handled.alt(identity))

    assert returns._errors_handled  # noqa: SLF001


@pytest.mark.parametrize(
    'container',
    [
        Success(1),
        Failure(1),
        IOSuccess(1),
        IOFailure(1),
        RequiresContextIOResult.from_value(1),
        RequiresContextIOResult.from_failure(1),
        RequiresContextFutureResult.from_value(1),
        RequiresContextFutureResult.from_failure(1),
        RequiresContextResult.from_value(1),
        RequiresContextResult.from_failure(1),
    ],
)
def test_error_not_handled(returns: ReturnsAsserts, container):
    """Demo on how to use ``pytest`` helpers to work with error handling."""
    assert not returns._errors_handled  # noqa: SLF001
    error_handled = _under_test(container)

    assert not returns.is_error_handled(container)
    assert not returns.is_error_handled(error_handled)
    assert not returns.is_error_handled(error_handled.map(identity))
    assert not returns.is_error_handled(error_handled.alt(identity))

    assert not returns._errors_handled  # noqa: SLF001


@pytest.mark.anyio
@pytest.mark.parametrize(
    'container',
    [
        FutureResult.from_value(1),
        FutureResult.from_failure(1),
        RequiresContextFutureResult.from_value(1),
        RequiresContextFutureResult.from_failure(1),
    ],
)
async def test_error_not_handled_async(returns: ReturnsAsserts, container):
    """Demo on how to use ``pytest`` helpers to work with error handling."""
    error_handled = _under_test(container)

    assert not returns.is_error_handled(container)
    assert not returns.is_error_handled(error_handled)
    assert not returns.is_error_handled(error_handled.map(identity))
    assert not returns.is_error_handled(error_handled.alt(identity))



from functools import reduce
from typing import TypeVar

_InstanceType = TypeVar('_InstanceType')
_PipelineStepType = TypeVar('_PipelineStepType')
_ReturnType = TypeVar('_ReturnType')


def flow(
    instance: _InstanceType,
    *functions: _PipelineStepType,
) -> _ReturnType:  # type: ignore[type-var]
    """
    Allows to compose a value and up to multiple functions that use this value.

    All starts with the value itself.
    Each next function uses the previous result as an input parameter.

    We use a custom ``mypy`` plugin to make sure types are correct.
    Otherwise, it is currently impossible to properly type this function.

    Currently, ``flow`` has a hard limit of 21 steps.
    Because, it is not possible to type it otherwise.
    We need a hard limit.
    See: https://github.com/dry-python/returns/issues/461

    Here's how it should be used:

    .. code:: python

       >>> from returns.pipeline import flow

       >>> # => executes: str(float(int('1')))
       >>> assert flow('1', int, float, str) == '1.0'

    This function is closely related
    to :func:`pipe <returns._internal.pipeline.pipe.pipe>`:

    .. code:: python

      >>> from returns.pipeline import pipe
      >>> assert flow('1', int, float, str) == pipe(int, float, str)('1')

    See also:
        - https://stackoverflow.com/a/41585450/4842742
        - https://github.com/gcanti/fp-ts/blob/master/src/pipeable.ts

    Requires our :ref:`mypy plugin <mypy-plugins>`.
    """
    return reduce(  # type: ignore
        lambda composed, function: function(composed),  # type: ignore
        functions,
        instance,
    )
